import json
from abc import abstractmethod
from typing import Any, Dict


class SchemaError(Exception):
    '''Custom Schema Error'''

    pass


class BaseSchema:
    _json_schema: Dict[str, Any]

    # def validate(self, obj: Any) -> None:
    #     pass

    @property
    @abstractmethod
    def json(self) -> Any:  # pragma: no cover
        pass

    def to_json(self, **dumps_kws) -> str:
        '''Generate a JSON string representation of this schema

        Parameters
        ----------
        dumps_kws : dict
            Parameters to pass to ``json.dumps``.

        Returns
        -------
        str
        '''
        return json.dumps(self.json, **dumps_kws)

    @classmethod
    @abstractmethod
    def from_json(cls, obj):  # pragma: no cover
        pass



from typing import Dict, Tuple, Union

from numpy.typing import DTypeLike  # noqa: F401

DimsT = Tuple[Union[str, None]]
ShapeT = Tuple[Union[int, None]]
ChunksT = Union[bool, Dict[str, Union[int, None]]]



import dask.array
import jsonschema
import numpy as np
import pytest
import xarray as xr

from xarrera import DataArraySchema, DatasetSchema
from xarrera.base import SchemaError
from xarrera.components import (
    ArrayTypeSchema,
    AttrSchema,
    AttrsSchema,
    ChunksSchema,
    DimsSchema,
    DTypeSchema,
    NameSchema,
    ShapeSchema,
)
from xarrera.dataarray import CoordsSchema


@pytest.fixture
def ds():
    ds = xr.Dataset(
        {
            'x': xr.DataArray(np.arange(4) - 2, dims='x'),
            'foo': xr.DataArray(np.ones(4, dtype='i4'), dims='x'),
            'bar': xr.DataArray(np.arange(8, dtype=np.float32).reshape(4, 2), dims=('x', 'y')),
        }
    )
    return ds


@pytest.mark.parametrize(
    'component, schema_args, validate, json',
    [
        (DTypeSchema, np.integer, ['i4', 'int', np.int32], 'integer'),
        (DTypeSchema, np.int64, ['i8', np.int64], '<i8'),
        (DTypeSchema, '<i8', ['i8', np.int64], '<i8'),
        (DTypeSchema, np.number, [np.int32, np.float64, np.complex128], 'number'),
        (DTypeSchema, np.floating, [np.float32, np.float64], 'floating'),
        (DTypeSchema, np.complexfloating, [np.complex64, np.complex128], 'complexfloating'),
        (DTypeSchema, np.inexact, [np.float32, np.complex128], 'inexact'),
        (DTypeSchema, np.character, [np.bytes_, np.str_], 'character'),
        (DimsSchema, ('foo', None), [('foo', 'bar'), ('foo', 'baz')], ['foo', None]),
        (DimsSchema, ('foo', 'bar'), [('foo', 'bar')], ['foo', 'bar']),
        (ShapeSchema, (1, 2, None), [(1, 2, 3), (1, 2, 5)], [1, 2, None]),
        (ShapeSchema, (1, 2, 3), [(1, 2, 3)], [1, 2, 3]),
        (NameSchema, 'foo', ['foo'], 'foo'),
        (ArrayTypeSchema, np.ndarray, [np.array([1, 2, 3])], "<class 'numpy.ndarray'>"),
        (
            ArrayTypeSchema,
            dask.array.Array,
            [dask.array.zeros(4)],
            "<class 'dask.array.core.Array'>",
        ),
        # schema_args for ChunksSchema include [chunks, dims, shape]
        (ChunksSchema, True, [(((1, 1),), ('x',), (2,))], True),
        (ChunksSchema, {'x': 2}, [(((2, 2),), ('x',), (4,))], {'x': 2}),
        (ChunksSchema, {'x': (2, 2)}, [(((2, 2),), ('x',), (4,))], {'x': [2, 2]}),
        (ChunksSchema, {'x': [2, 2]}, [(((2, 2),), ('x',), (4,))], {'x': [2, 2]}),
        (ChunksSchema, {'x': 4}, [(((4,),), ('x',), (4,))], {'x': 4}),
        (ChunksSchema, {'x': -1}, [(((4,),), ('x',), (4,))], {'x': -1}),
        (ChunksSchema, {'x': (1, 2, 1)}, [(((1, 2, 1),), ('x',), (4,))], {'x': [1, 2, 1]}),
        (
            ChunksSchema,
            {'x': 2, 'y': -1},
            [(((2, 2), (10,)), ('x', 'y'), (4, 10))],
            {'x': 2, 'y': -1},
        ),
        (
            AttrsSchema,
            {'foo': AttrSchema(value='bar')},
            [{'foo': 'bar'}],
            {
                'allow_extra_keys': True,
                'require_all_keys': True,
                'attrs': {'foo': {'type': None, 'value': 'bar'}},
            },
        ),
        (
            AttrsSchema,
            {'foo': AttrSchema(value=1)},
            [{'foo': 1}],
            {
                'allow_extra_keys': True,
                'require_all_keys': True,
                'attrs': {'foo': {'type': None, 'value': 1}},
            },
        ),
        (
            CoordsSchema,
            {'x': DataArraySchema(name='x')},
            [{'x': xr.DataArray([0, 1], name='x')}],
            {'coords': {'x': {'name': 'x'}}, 'allow_extra_keys': True, 'require_all_keys': True},
        ),
    ],
)
def test_component_schema(component, schema_args, validate, json):
    schema = component(schema_args)
    for v in validate:
        if component in [ChunksSchema]:  # special case construction
            schema.validate(*v)
        else:
            schema.validate(v)
    assert schema.json == json
    assert isinstance(schema.to_json(), str)

    # validate schema
    jsonschema.validate(schema.json, schema._json_schema)

    # json roundtrip
    component.from_json(schema.json).json == json


@pytest.mark.parametrize(
    'type, value, validate, json',
    [
        (str, None, 'foo', {'type': str, 'value': None}),
        (None, 'foo', 'foo', {'type': None, 'value': 'foo'}),
        (str, 'foo', 'foo', {'type': str, 'value': 'foo'}),
    ],
)
def test_attr_schema(type, value, validate, json):
    schema = AttrSchema(type=type, value=value)
    schema.validate(validate)
    assert schema.json == json
    # assert isinstance(schema.to_json(), str)


@pytest.mark.parametrize(
    'component, schema_args, validate, match',
    [
        (DTypeSchema, np.integer, np.float32, r'.*float.*'),
        (DTypeSchema, np.number, np.str_, r'.*str.*'),
        (DTypeSchema, np.floating, np.int32, r'.*int32.*'),
        (DTypeSchema, np.complexfloating, np.float64, r'.*float64.*'),
        (DimsSchema, ('foo', 'bar'), ('foo',), r'.*length.*'),
        (DimsSchema, ('foo', 'bar'), ('foo', 'baz'), r'.*mismatch.*'),
        (ShapeSchema, (1, 2, None), (1, 2), r'.*number of dimensions.*'),
        (ShapeSchema, (1, 4, 4), (1, 3, 4), r'.*mismatch.*'),
        (NameSchema, 'foo', 'bar', r'.*name bar != foo.*'),
        (ArrayTypeSchema, np.ndarray, 'bar', r'.*array_type.*'),
        # schema_args for ChunksSchema include [chunks, dims, shape]
        (ChunksSchema, {'x': 3}, (((2, 2),), ('x',), (4,)), r'.*(3).*'),
        (ChunksSchema, {'x': (2, 1)}, (((2, 2),), ('x',), (4,)), r'.*(2, 1).*'),
        (ChunksSchema, {'x': (2, 1)}, (None, ('x',), (4,)), r'.*expected array to be chunked.*'),
        (ChunksSchema, True, (None, ('x',), (4,)), r'.*expected array to be chunked.*'),
        (
            ChunksSchema,
            False,
            (((2, 2),), ('x',), (4,)),
            r'.*expected unchunked array but it is chunked*',
        ),
        (ChunksSchema, {'x': -1}, (((1, 2, 1),), ('x',), (4,)), r'.*did not match.*'),
        (ChunksSchema, {'x': 2}, (((2, 3, 2),), ('x',), (7,)), r'.*did not match.*'),
        (ChunksSchema, {'x': 2}, (((2, 2, 3),), ('x',), (7,)), r'.*did not match.*'),
        (ChunksSchema, {'x': 2, 'y': -1}, (((2, 2), (5, 5)), ('x', 'y'), (4, 10)), r'.*(5).*'),
    ],
)
def test_component_raises_schema_error(component, schema_args, validate, match):
    schema = component(schema_args)
    with pytest.raises(SchemaError, match=match):
        if component in [ChunksSchema]:  # special case construction
            schema.validate(*validate)
        else:
            schema.validate(validate)


def test_chunks_schema_raises_for_invalid_chunks():
    with pytest.raises(ValueError, match=r'.*int.*'):
        schema = ChunksSchema(chunks=2)
        schema.validate(((2, 2),), ('x',), (4,))


def test_unknown_array_type_raises():
    with pytest.raises(ValueError, match=r'.*unknown array_type.*'):
        _ = ArrayTypeSchema.from_json('foo.array')


def test_dataarray_empty_constructor():
    da = xr.DataArray(np.ones(4, dtype='i4'))
    da_schema = DataArraySchema()
    assert hasattr(da_schema, 'validate')
    jsonschema.validate(da_schema.json, da_schema._json_schema)
    assert da_schema.json == {}
    da_schema.validate(da)


@pytest.mark.parametrize(
    'kind, component, schema_args',
    [
        ('dtype', DTypeSchema, 'i4'),
        ('dims', DimsSchema, ('x', None)),
        ('shape', ShapeSchema, (2, None)),
        ('name', NameSchema, 'foo'),
        ('array_type', ArrayTypeSchema, np.ndarray),
        ('chunks', ChunksSchema, False),
    ],
)
def test_dataarray_component_constructors(kind, component, schema_args):
    da = xr.DataArray(np.zeros((2, 4), dtype='i4'), dims=('x', 'y'), name='foo')
    comp_schema = component(schema_args)
    schema = DataArraySchema(**{kind: schema_args})
    assert comp_schema.json == getattr(schema, kind).json
    jsonschema.validate(schema.json, schema._json_schema)
    assert isinstance(getattr(schema, kind), component)

    # json roundtrip
    rt_schema = DataArraySchema.from_json(schema.json)
    assert isinstance(rt_schema, DataArraySchema)
    assert rt_schema.json == schema.json

    schema.validate(da)


def test_dataarray_schema_validate_raises_for_invalid_input_type():
    ds = xr.Dataset()
    schema = DataArraySchema()
    with pytest.raises(ValueError, match='Input must be a xarray.DataArray'):
        schema.validate(ds)


def test_dataset_empty_constructor():
    ds_schema = DatasetSchema()
    assert hasattr(ds_schema, 'validate')
    jsonschema.validate(ds_schema.json, ds_schema._json_schema)
    ds_schema.json == {}


def test_dataset_example(ds):
    ds_schema = DatasetSchema(
        {
            'foo': DataArraySchema(name='foo', dtype=np.int32, dims=['x']),
            'bar': DataArraySchema(name='bar', dtype=np.floating, dims=['x', 'y']),
        }
    )

    jsonschema.validate(ds_schema.json, ds_schema._json_schema)

    assert list(ds_schema.json['data_vars'].keys()) == ['foo', 'bar']
    ds_schema.validate(ds)

    ds['foo'] = ds.foo.astype('float32')
    with pytest.raises(SchemaError, match='dtype'):
        ds_schema.validate(ds)

    ds = ds.drop_vars('foo')
    with pytest.raises(SchemaError, match='variable foo'):
        ds_schema.validate(ds)

    # json roundtrip
    rt_schema = DatasetSchema.from_json(ds_schema.json)
    assert isinstance(rt_schema, DatasetSchema)
    rt_schema.json == ds_schema.json


def test_checks_ds(ds):
    def check_foo(ds):
        assert 'foo' in ds

    ds_schema = DatasetSchema(checks=[check_foo])
    ds_schema.validate(ds)

    ds = ds.drop_vars('foo')
    with pytest.raises(AssertionError):
        ds_schema.validate(ds)

    ds_schema = DatasetSchema(checks=[])
    ds_schema.validate(ds)

    # TODO
    # with pytest.raises(ValueError):
    #     DatasetSchema(checks=[2])


def test_dataset_with_attrs_schema():
    name = 'name'
    expected_value = 'expected_value'
    actual_value = 'actual_value'
    ds = xr.Dataset(attrs={name: actual_value})
    ds_schema = DatasetSchema(attrs={name: AttrSchema(value=expected_value)})
    jsonschema.validate(ds_schema.json, ds_schema._json_schema)

    ds_schema_2 = DatasetSchema(attrs=AttrsSchema({name: AttrSchema(value=expected_value)}))
    jsonschema.validate(ds_schema_2.json, ds_schema_2._json_schema)
    with pytest.raises(SchemaError):
        ds_schema.validate(ds)
    with pytest.raises(SchemaError):
        ds_schema_2.validate(ds)


def test_attrs_extra_key():
    name = 'name'
    value = 'value_2'
    name_2 = 'name_2'
    value_2 = 'value_2'
    ds = xr.Dataset(attrs={name: value})
    ds_schema = DatasetSchema(
        attrs=AttrsSchema(
            attrs={
                name: AttrSchema(
                    value=value,
                ),
                name_2: AttrSchema(value=value_2),
            },
            require_all_keys=True,
        )
    )
    jsonschema.validate(ds_schema.json, ds_schema._json_schema)

    with pytest.raises(SchemaError):
        ds_schema.validate(ds)


def test_attrs_missing_key():
    name = 'name'
    value = 'value_2'
    name_2 = 'name_2'
    value_2 = 'value_2'
    ds = xr.Dataset(attrs={name: value, name_2: value_2})
    ds_schema = DatasetSchema(
        attrs=AttrsSchema(attrs={name: AttrSchema(value=value)}, allow_extra_keys=False)
    )
    with pytest.raises(SchemaError):
        ds_schema.validate(ds)


def test_checks_da(ds):
    da = ds['foo']

    def check_foo(da):
        assert da.name == 'foo'

    def check_bar(da):
        assert da.name == 'bar'

    schema = DataArraySchema(checks=[check_foo])
    schema.validate(da)

    schema = DataArraySchema(checks=[check_bar])
    with pytest.raises(AssertionError):
        schema.validate(da)

    schema = DataArraySchema(checks=[])
    schema.validate(da)

    with pytest.raises(ValueError):
        DataArraySchema(checks=[2])



from invoke import task  # pragma: no cover

SRC_DIR = 'xarrera'  # pragma: no cover
TEST_DIR = 'tests'  # pragma: no cover


@task  # pragma: no cover
def mypy(c):  # pragma: no cover
    c.run(f'mypy {SRC_DIR} {TEST_DIR}')  # pragma: no cover



from collections.abc import Iterable
from typing import Any, Dict, Hashable, Mapping, Optional, Tuple, Union

import numpy as np

from .base import BaseSchema, SchemaError
from .types import ChunksT, DimsT, DTypeLike, ShapeT


class DTypeSchema(BaseSchema):
    '''Datatype schema

    Parameters
    ----------
    dtype : DTypeLike
        Datatype definition, may be (string, np.dtype, etc.)

    Raises
    ------
    SchemaError
    '''

    _NP_ABSTRACT_TYPES = (
        np.generic,
        np.number,
        np.integer,
        np.signedinteger,
        np.unsignedinteger,
        np.inexact,
        np.floating,
        np.complexfloating,
        np.flexible,
        np.character,
    )

    _NP_ABSTRACT_TYPES_STR = (
        'generic',
        'number',
        'integer',
        'signedinteger',
        'unsignedinteger',
        'inexact',
        'floating',
        'complexfloating',
        'flexible',
        'character',
    )

    _json_schema = {'type': 'string'}

    def __init__(self, dtype: DTypeLike) -> None:
        if dtype in type(self)._NP_ABSTRACT_TYPES:
            self.dtype = dtype
        else:
            self.dtype = np.dtype(dtype)

    @classmethod
    def from_json(cls, obj: str):
        if obj in cls._NP_ABSTRACT_TYPES_STR:
            dtype = getattr(np, obj)
        else:
            dtype = obj
        return cls(dtype)

    def validate(self, dtype: DTypeLike) -> None:
        '''Validate dtype

        Parameters
        ----------
        dtype : Any
            Dtype of the DataArray.
        '''
        if not np.issubdtype(dtype, self.dtype):
            raise SchemaError(f'dtype {dtype} != {self.dtype}')

    @property
    def json(self) -> str:
        if isinstance(self.dtype, np.dtype):
            return self.dtype.str
        else:
            # fallbacks
            return str(getattr(self.dtype, '__name__', str(self.dtype)))


class DimsSchema(BaseSchema):
    '''Dimensions schema

    Parameters
    ----------
    dims : str or iterable of str
        Dimensions definition, ``None`` may be used as a wildcard.

    Raises
    ------
    SchemaError
    '''

    _json_schema = {'type': 'array', 'items': {'type': ['string', 'null']}}

    def __init__(self, dims: DimsT) -> None:
        self.dims = dims

    @classmethod
    def from_json(cls, obj: DimsT):
        return cls(obj)

    def validate(self, dims: tuple) -> None:
        '''Validate dimensions

        Parameters
        ----------
        dims : Tuple[Union[str, None]]
            Dimensions of the DataArray. `None` may be used as a wildcard value.
        '''
        if len(self.dims) != len(dims):
            raise SchemaError(f'length of dims does not match: {len(dims)} != {len(self.dims)}')

        for i, (actual, expected) in enumerate(zip(dims, self.dims)):
            if expected is not None and actual != expected:
                raise SchemaError(f'dim mismatch in axis {i}: {actual} != {expected}')

    @property
    def json(self) -> list:
        return list(self.dims)


class ShapeSchema(BaseSchema):
    '''Shape schema

    Parameters
    ----------
    shape : iterable of ints
        Shape definition, ``None`` may be used as a wildcard.

    Raises
    ------
    SchemaError
    '''

    _json_schema = {'type': 'array'}

    def __init__(self, shape: ShapeT) -> None:
        self.shape = shape

    @classmethod
    def from_json(cls, obj: ShapeT):
        return cls(obj)

    def validate(self, shape: tuple) -> None:
        '''Validate shape

        Parameters
        ----------
        shape : ShapeT
            Shape of the DataArray. `None` may be used as a wildcard value.
        '''
        if len(self.shape) != len(shape):
            raise SchemaError(
                f'number of dimensions in shape ({len(shape)}) != da.ndim ({len(self.shape)})'
            )

        for i, (actual, expected) in enumerate(zip(shape, self.shape)):
            if expected is not None and actual != expected:
                raise SchemaError(f'shape mismatch in axis {i}: {actual} != {expected}')

    @property
    def json(self) -> list:
        return list(self.shape)


class NameSchema(BaseSchema):
    '''Name schema

    Parameters
    ----------
    name : str
        Name definition.

    Raises
    ------
    SchemaError
    '''

    _json_schema = {'type': 'string'}

    def __init__(self, name: str) -> None:
        self.name = name

    @classmethod
    def from_json(cls, obj: str):
        return cls(obj)

    def validate(self, name: Hashable) -> None:
        '''Validate name

        Parameters
        ----------
        name : str, optional
            Name of the DataArray. Currently requires an exact string match.
        '''
        # TODO: support regular expressions
        # - http://json-schema.org/understanding-json-schema/reference/regular_expressions.html
        # - https://docs.python.org/3.9/library/re.html
        if self.name != name:
            raise SchemaError(f'name {name} != {self.name}')

    @property
    def json(self) -> str:
        return self.name


class ChunksSchema(BaseSchema):
    '''Chunks schema

    Parameters
    ----------
    chunks : dict or bool
        Chunks definition. If ``bool``, whether validated object should be chunked.
        If ``dict``, mapping of dimension name to chunk size. None may be used as a wildcard.

    Raises
    ------
    SchemaError
    '''

    _json_schema = {'type': ['boolean', 'object']}

    def __init__(self, chunks: ChunksT) -> None:
        self.chunks = chunks

    @classmethod
    def from_json(cls, obj: dict):
        return cls(obj)  # TODO: this will likely need input validation.

    def validate(
        self, chunks: Optional[Tuple[Tuple[int, ...], ...]], dims: Tuple, shape: Tuple[int, ...]
    ) -> None:
        '''Validate chunks

        Parameters
        ----------
        chunks : tuple
            Chunks from ``DataArray.chunks``
        dims : tuple of str
            Dimension keys from array.
        shape : tuple of int
            Shape of array.
        '''

        if isinstance(self.chunks, bool):
            if self.chunks and not chunks:
                raise SchemaError('expected array to be chunked but it is not')
            elif not self.chunks and chunks:
                raise SchemaError('expected unchunked array but it is chunked')
        elif isinstance(self.chunks, dict):
            if chunks is None:
                raise SchemaError('expected array to be chunked but it is not')
            dim_chunks = dict(zip(dims, chunks))
            dim_sizes = dict(zip(dims, shape))
            # check whether chunk sizes are regular because we assume the first chunk to be representative below
            for key, ec in self.chunks.items():
                if isinstance(ec, int):
                    # handles case of expected chunksize is shorthand of -1 which translates to the full length of dimension
                    if ec < 0:
                        ec = dim_sizes[key]
                    ac = dim_chunks[key]
                    if any([a != ec for a in ac[:-1]]) or ac[-1] > ec:
                        raise SchemaError(f'{key} chunks did not match: {ac} != {ec}')

                else:  # assumes ec is an iterable
                    ac = dim_chunks[key]
                    if ec is not None and tuple(ac) != tuple(ec):
                        raise SchemaError(f'{key} chunks did not match: {ac} != {ec}')
        else:
            raise ValueError(f'got unknown chunks type: {type(self.chunks)}')

    @property
    def json(self) -> Union[bool, Dict[str, Any]]:
        if isinstance(self.chunks, bool):
            return self.chunks
        else:
            obj = {}
            for key, val in self.chunks.items():
                if isinstance(val, Iterable):
                    obj[key] = list(val)
                else:
                    obj[key] = val
            return obj


class ArrayTypeSchema(BaseSchema):
    '''Array type schema

    Parameters
    ----------
    array_type : str or object
        Array type definition.

    Raises
    ------
    SchemaError
    '''

    _json_schema = {'type': 'string'}

    def __init__(self, array_type: Any) -> None:
        self.array_type = array_type

    @classmethod
    def from_json(cls, obj: str):
        array_type: Any  # TODO: figure out how to optionally include the dask array type

        if obj == "<class 'dask.array.core.Array'>":
            import dask.array as da

            array_type = da.Array
        elif obj == "<class 'numpy.ndarray'>":
            array_type = np.ndarray
        else:
            raise ValueError(f'unknown array_type: {obj}')
        return cls(array_type)

    def validate(self, array: Any) -> None:
        '''Validate array_type

        Parameters
        ----------
        array : array_like
            array_type of the DataArray. `None` may be used as a wildcard value.
        '''
        if not isinstance(array, self.array_type):
            raise SchemaError(f'array_type {type(array)} != {self.array_type}')

    @property
    def json(self) -> str:
        return str(self.array_type)


class AttrSchema(BaseSchema):
    '''Attribute schema

    Parameters
    ----------
    type : object
        Attribute type definition.
    value :
        Attribute value definition.

    Raises
    ------
    SchemaError
    '''

    _json_schema = {'type': 'string', 'value': ['string', 'number', 'array', 'boolean', 'null']}

    def __init__(self, type: Any = None, value: Any = None):
        self.type = type
        self.value = value

    @classmethod
    def from_json(cls, obj: str):
        return cls(obj)

    def validate(self, attr: Any):
        '''Validate attrs

        Parameters
        ----------
        attr : any
            attribute, `None` may be used as a wildcard value.
        '''
        if self.type is not None:
            if not isinstance(attr, self.type):
                raise SchemaError(f'attrs {attr} is not of type {self.type}')

        if self.value is not None:
            if self.value is not None and self.value != attr:
                raise SchemaError(f'name {attr} != {self.value}')

    @property
    def json(self) -> dict:
        return {'type': self.type, 'value': self.value}


class AttrsSchema(BaseSchema):
    '''Attributes schema

    Parameters
    ----------
    attrs : str or iterable of str
        Attributes definition
    require_all_keys : bool
        Whether require to all coordinates included in ``attrs``
    allow_extra_keys : bool
        Whether to allow coordinates not included in ``attrs`` dict

    Raises
    ------
    SchemaError
    '''

    _json_schema = {
        'type': 'object',
        'properties': {
            'require_all_keys': {
                'type': 'boolean'
            },  # Question: is this the same as JSON's additionalProperties?
            'allow_extra_keys': {'type': 'boolean'},
            'attrs': {'type': 'object'},
        },
    }

    def __init__(
        self,
        attrs: Mapping[Hashable, AttrSchema],
        require_all_keys: bool = True,
        allow_extra_keys: bool = True,
    ) -> None:
        self.attrs = attrs
        self.require_all_keys = require_all_keys
        self.allow_extra_keys = allow_extra_keys

    @classmethod
    def from_json(cls, obj: dict):
        attrs = {}
        for key, val in obj['attrs'].items():
            attrs[key] = AttrSchema(**val)
        return cls(
            attrs,
            require_all_keys=obj['require_all_keys'],
            allow_extra_keys=obj['allow_extra_keys'],
        )

    def validate(self, attrs: Any) -> None:
        '''Validate attrs

        Parameters
        ----------
        attrs : dict_like
            attrs dict, `None` may be used as a wildcard value.
        '''

        if self.require_all_keys:
            missing_keys = set(self.attrs) - set(attrs)
            if missing_keys:
                raise SchemaError(f'attrs has missing keys: {missing_keys}')

        if not self.allow_extra_keys:
            extra_keys = set(attrs) - set(self.attrs)
            if extra_keys:
                raise SchemaError(f'attrs has extra keys: {extra_keys}')

        for key, attr_schema in self.attrs.items():
            if key not in attrs:
                raise SchemaError(f'key {key} not in attrs')
            else:
                attr_schema.validate(attrs[key])

    @property
    def json(self) -> dict:
        obj = {
            'require_all_keys': self.require_all_keys,
            'allow_extra_keys': self.allow_extra_keys,
            'attrs': {k: v.json for k, v in self.attrs.items()},
        }
        return obj



from typing import Any, Callable, Dict, Hashable, Iterable, Optional, Union

import xarray as xr

from .base import BaseSchema, SchemaError
from .components import AttrSchema, AttrsSchema
from .dataarray import CoordsSchema, DataArraySchema


class DatasetSchema(BaseSchema):
    '''A light-weight xarray.Dataset validator

    Parameters
    ----------
    data_vars : mapping of variable names and DataArraySchemas, optional
        Per-variable DataArraySchema's, by default None
    checks : Iterable[Callable], optional
        Dataset wide checks, by default None
    '''

    _json_schema = {
        'type': 'object',
        'properties': {
            'data_vars': {'type': 'object'},
            'coords': {'type': 'object'},
            'attrs': {'type': 'object'},
        },
    }

    def __init__(
        self,
        data_vars: Optional[Dict[Hashable, Optional[DataArraySchema]]] = None,
        coords: Union[CoordsSchema, Dict[Hashable, DataArraySchema], None] = None,
        attrs: Union[AttrsSchema, Dict[Hashable, AttrSchema], None] = None,
        checks: Optional[Iterable[Callable]] = None,
    ) -> None:
        self.data_vars = data_vars  # type: ignore
        self.coords = coords  # type: ignore
        self.attrs = attrs  # type: ignore
        self.checks = checks

    @classmethod
    def from_json(cls, obj: dict):
        kwargs = {}
        if 'data_vars' in obj:
            kwargs['data_vars'] = {
                k: DataArraySchema.from_json(v) for k, v in obj['data_vars'].items()
            }
        if 'coords' in obj:
            kwargs['coords'] = {k: CoordsSchema.from_json(v) for k, v in obj['coords'].items()}
        if 'attrs' in obj:
            kwargs['attrs'] = {k: AttrsSchema.from_json(v) for k, v in obj['attrs'].items()}

        return cls(**kwargs)

    def validate(self, ds: xr.Dataset) -> None:
        '''Check if the Dataset complies with the Schema.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset to be validated

        Returns
        -------
        xr.Dataset
            Validated Dataset

        Raises
        ------
        SchemaError
        '''

        if self.data_vars is not None:
            for key, da_schema in self.data_vars.items():
                if da_schema is not None:
                    if key not in ds.data_vars:
                        raise SchemaError(f'data variable {key} not in ds')
                    else:
                        da_schema.validate(ds.data_vars[key])

        if self.coords is not None:  # pragma: no cover
            raise NotImplementedError('coords schema not implemented yet')

        if self.attrs:
            self.attrs.validate(ds.attrs)

        if self.checks:
            for check in self.checks:
                check(ds)

    @property
    def attrs(self) -> Union[AttrsSchema, None]:
        return self._attrs

    @attrs.setter
    def attrs(self, value: Union[AttrsSchema, Dict[Hashable, Any], None]):
        if value is None or isinstance(value, AttrsSchema):
            self._attrs = value
        else:
            self._attrs = AttrsSchema(value)

    @property
    def data_vars(self) -> Optional[Dict[Hashable, Optional[DataArraySchema]]]:
        return self._data_vars  # type: ignore

    @data_vars.setter
    def data_vars(self, value: Optional[Dict[Hashable, Optional[DataArraySchema]]]):
        if isinstance(value, dict):
            self._data_vars = {}
            for k, v in value.items():
                if isinstance(v, DataArraySchema):
                    self._data_vars[k] = v
                else:
                    self._data_vars[k] = DataArraySchema(**v)  # type: ignore
        elif value is None:
            self._data_vars = None  # type: ignore
        else:
            raise ValueError('must set data_vars with a dict')

    @property
    def coords(self) -> Optional[CoordsSchema]:
        return self._coords  # type: ignore

    @coords.setter
    def coords(self, value: Optional[Union[CoordsSchema, Dict[Hashable, DataArraySchema]]]):
        if value is None or isinstance(value, CoordsSchema):
            self._coords = value
        else:
            self._coords = CoordsSchema(value)

    @property
    def json(self):
        obj = {'data_vars': {}, 'attrs': self.attrs.json if self.attrs is not None else {}}
        if self.data_vars:
            for key, var in self.data_vars.items():
                obj['data_vars'][key] = var.json
        if self.coords:
            obj['coords'] = self.coords.json
        return obj



from __future__ import annotations

from typing import Any, Callable, Dict, Hashable, List, Mapping, Optional, Union

import xarray as xr

from .base import BaseSchema, SchemaError
from .components import (
    ArrayTypeSchema,
    AttrsSchema,
    ChunksSchema,
    DimsSchema,
    DTypeSchema,
    NameSchema,
    ShapeSchema,
)
from .types import ChunksT, DimsT, DTypeLike, ShapeT


class DataArraySchema(BaseSchema):
    '''A light-weight xarray.DataArray validator

    Parameters
    ----------
    dtype : DTypeLike or DTypeSchema, optional
        Datatype of the the variable. If a string is specified it must be a valid NumPy data type value, by default None
    shape : ShapeT or ShapeSchema, optional
        Shape of the DataArray. `None` may be used as a wildcard value. By default None
    dims : DimsT or DimsSchema, optional
        Dimensions of the DataArray.  `None` may be used as a wildcard value. By default None
    chunks : Union[bool, Dict[str, Union[int, None]]], optional
        If bool, specifies whether DataArray is chunked or not, agnostic to chunk sizes.
        If dict, includes the expected chunks for the DataArray, by default None
    name : str, optional
        Name of the DataArray, by default None
    array_type : Any, optional
        Type of the underlying data in a DataArray (e.g. `numpy.ndarray`), by default None
    checks : List[Callable], optional
        List of callables that take and return a DataArray, by default None
    '''

    _json_schema = {'type': 'object'}
    _schema_slots = ['dtype', 'dims', 'shape', 'coords', 'name', 'chunks', 'attrs', 'array_type']

    _dtype: Union[DTypeSchema, None]
    _shape: Union[ShapeSchema, None]
    _dims: Union[DimsSchema, None]
    _name: Union[NameSchema, None]
    _coords: Union[Any, None]
    _chunks: Union[ChunksSchema, None]
    _attrs: Union[AttrsSchema, None]
    _array_type: Union[ArrayTypeSchema, None]

    def __init__(
        self,
        dtype: Union[DTypeLike, DTypeSchema, None] = None,
        shape: Union[ShapeT, ShapeSchema, None] = None,
        dims: Union[DimsT, DimsSchema, None] = None,
        name: Union[str, NameSchema, None] = None,
        coords: Optional[Dict[str, Any]] = None,
        chunks: Union[ChunksT, ChunksSchema, None] = None,
        array_type: Any = None,
        attrs: Optional[Mapping[str, Any]] = None,
        checks: Optional[List[Callable]] = None,
    ) -> None:
        # see https://github.com/python/mypy/issues/3004
        self.dtype = dtype  # type: ignore
        self.shape = shape  # type: ignore
        self.dims = dims  # type: ignore
        self.name = name  # type: ignore
        self.coords = coords  # type: ignore
        self.chunks = chunks  # type: ignore
        self.attrs = attrs  # type: ignore
        self.array_type = array_type  # type: ignore
        self.checks = checks  # type: ignore

    @property
    def dtype(self) -> Union[DTypeSchema, None]:
        return self._dtype

    @dtype.setter
    def dtype(self, value: Union[DTypeSchema, DTypeLike, None]):
        if value is None or isinstance(value, DTypeSchema):
            self._dtype = value
        else:
            self._dtype = DTypeSchema(value)

    @property
    def dims(self) -> Union[DimsSchema, None]:
        return self._dims

    @dims.setter
    def dims(self, value):
        if value is None or isinstance(value, DimsSchema):
            self._dims = value
        else:
            self._dims = DimsSchema(value)

    @property
    def shape(self) -> Optional[ShapeSchema]:
        return self._shape

    @shape.setter
    def shape(self, value: Union[ShapeSchema, ShapeT, None]):
        if value is None or isinstance(value, ShapeSchema):
            self._shape = value
        else:
            self._shape = ShapeSchema(value)

    @property
    def chunks(self) -> Optional[ChunksSchema]:
        return self._chunks

    @chunks.setter
    def chunks(self, value):
        if value is None or isinstance(value, ChunksSchema):
            self._chunks = value
        else:
            self._chunks = ChunksSchema(value)

    @property
    def name(self) -> Optional[NameSchema]:
        return self._name

    @name.setter
    def name(self, value):
        if value is None or isinstance(value, NameSchema):
            self._name = value
        else:
            self._name = NameSchema(value)

    @property
    def array_type(self) -> Optional[ArrayTypeSchema]:
        return self._array_type

    @array_type.setter
    def array_type(self, value):
        if value is None or isinstance(value, ArrayTypeSchema):
            self._array_type = value
        else:
            self._array_type = ArrayTypeSchema(value)

    @property
    def attrs(self) -> Optional[AttrsSchema]:
        return self._attrs

    @attrs.setter
    def attrs(self, value):
        if value is None or isinstance(value, AttrsSchema):
            self._attrs = value
        else:
            self._attrs = AttrsSchema(value)

    @property
    def coords(self) -> Optional[CoordsSchema]:
        return self._coords

    @coords.setter
    def coords(self, value):
        if value is None or isinstance(value, CoordsSchema):
            self._coords = value
        else:
            self._coords = CoordsSchema(value)

    @property
    def checks(self) -> List[Callable]:
        return self._checks

    @checks.setter
    def checks(self, value):
        if value is not None:
            if not all([callable(f) for f in value]):
                raise ValueError('All checks must be callables')
            self._checks = value
        else:
            self._checks = []

    def validate(self, da: xr.DataArray) -> None:
        '''Check if the DataArray complies with the Schema.

        Parameters
        ----------
        da : xr.DataArray
            DataArray to be validated

        Returns
        -------
        xr.DataArray
            Validated DataArray

        Raises
        ------
        SchemaError
        '''
        if not isinstance(da, xr.DataArray):
            raise ValueError('Input must be a xarray.DataArray')

        if self.dtype is not None:
            self.dtype.validate(da.dtype)

        if self.name is not None:
            self.name.validate(da.name)

        if self.dims is not None:
            self.dims.validate(da.dims)

        if self.shape is not None:
            self.shape.validate(da.shape)

        if self.coords is not None:
            self.coords.validate(da.coords)

        if self.chunks is not None:
            self.chunks.validate(da.chunks, da.dims, da.shape)

        if self.attrs:
            self.attrs.validate(da.attrs)

        if self.array_type is not None:
            self.array_type.validate(da.data)

        for check in self.checks:
            check(da)

    @property
    def json(self) -> dict:
        obj = {}
        for slot in self._schema_slots:
            try:
                obj[slot] = getattr(self, slot).json
            except AttributeError:
                pass
        return obj

    @classmethod
    def from_json(cls, obj: dict):
        kwargs = {}

        if 'dtype' in obj:
            kwargs['dtype'] = DTypeSchema.from_json(obj['dtype'])
        if 'shape' in obj:
            kwargs['shape'] = ShapeSchema.from_json(obj['shape'])
        if 'dims' in obj:
            kwargs['dims'] = DimsSchema.from_json(obj['dims'])
        if 'name' in obj:
            kwargs['name'] = NameSchema.from_json(obj['name'])
        if 'coords' in obj:
            kwargs['coords'] = CoordsSchema.from_json(obj['coords'])
        if 'chunks' in obj:
            kwargs['chunks'] = ChunksSchema.from_json(obj['chunks'])
        if 'array_type' in obj:
            kwargs['array_type'] = ArrayTypeSchema.from_json(obj['array_type'])
        if 'attrs' in obj:
            kwargs['attrs'] = AttrsSchema.from_json(obj['attrs'])

        return cls(**kwargs)


class CoordsSchema(BaseSchema):
    '''Schema container for Coordinates

    Parameters
    ----------
    coords : dict
        Dict of coordinate keys and ``DataArraySchema`` objects
    require_all_keys : bool
        Whether require to all coordinates included in ``coords``
    allow_extra_keys : bool
        Whether to allow coordinates not included in ``coords`` dict

    Raises
    ------
    SchemaError
    '''

    _json_schema = {
        'type': 'object',
        'properties': {
            'require_all_keys': {
                'type': 'boolean'
            },  # Question: is this the same as JSON's additionalProperties?
            'allow_extra_keys': {'type': 'boolean'},
            'coords': {'type': 'object'},
        },
    }

    def __init__(
        self,
        coords: Dict[Hashable, DataArraySchema],
        require_all_keys: bool = True,
        allow_extra_keys: bool = True,
    ) -> None:
        self.coords = coords
        self.require_all_keys = require_all_keys
        self.allow_extra_keys = allow_extra_keys

    @classmethod
    def from_json(cls, obj: dict):
        coords = obj.pop('coords', {})
        coords = {k: DataArraySchema(**v) for k, v in coords.items()}
        return cls(coords, **obj)

    def validate(self, coords: Any) -> None:
        '''Validate coords

        Parameters
        ----------
        coords : dict_like
            coords of the DataArray. `None` may be used as a wildcard value.
        '''

        if self.require_all_keys:
            missing_keys = set(self.coords) - set(coords)
            if missing_keys:
                raise SchemaError(f'coords has missing keys: {missing_keys}')

        if not self.allow_extra_keys:
            extra_keys = set(coords) - set(self.coords)
            if extra_keys:
                raise SchemaError(f'coords has extra keys: {extra_keys}')

        for key, da_schema in self.coords.items():
            if key not in coords:
                raise SchemaError(f'key {key} not in coords')
            else:
                da_schema.validate(coords[key])

    @property
    def json(self) -> dict:
        obj = {
            'require_all_keys': self.require_all_keys,
            'allow_extra_keys': self.allow_extra_keys,
            'coords': {k: v.json for k, v in self.coords.items()},
        }
        return obj



from ._version import __version__  # noqa: F401
from .base import SchemaError  # noqa: F401
from .components import (  # noqa: F401
    ArrayTypeSchema,
    AttrSchema,
    AttrsSchema,
    ChunksSchema,
    DimsSchema,
    DTypeSchema,
    NameSchema,
    ShapeSchema,
)
from .dataarray import CoordsSchema, DataArraySchema  # noqa: F401
from .dataset import DatasetSchema  # noqa: F401
